"""
Vector Store on Pinecone.
"""

from __future__ import annotations

__all__ = ["Pinecone"]

import copy
from typing import Any

from pinecone import NotFoundException
from pinecone import Pinecone as PineconeClient
from pinecone import PodSpec

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Pinecone(StoreProvider):
    api_key: str
    host: str | None
    index: str | None
    namespace: str | None
    content_metadata_field: str | None
    nparams: dict[str, Any]

    _database_client: PineconeClient
    _processor: ItemProcessor

    _collection_cache: dict[str, PineconeCollection]

    def __init__(
        self,
        api_key: str,
        host: str | None = None,
        index: str | None = None,
        namespace: str | None = None,
        content_metadata_field: str | None = "_content",
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            api_key:
                Pinecone API key.
            host:
                Pinecone host.
            index:
                Pinecone index mapped to vector store collection.
            namespace:
                Index namespace.
            content_metadata_field:
                Metadata field where content is stored.
                Pinecone doesn't support a separate content field.
                Defaults to "_content".
            nparams:
                Native parameters to pinecone client.
        """
        self.api_key = api_key
        self.host = host
        self.index = index
        self.namespace = namespace
        self.content_metadata_field = content_metadata_field
        self.nparams = nparams

        self._database_client = None
        self._processor = ItemProcessor()
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._database_client is not None:
            return

        self._database_client = PineconeClient(
            api_key=self.api_key,
            host=self.host,
            **self.nparams,
        )

    def _get_index_name(self, op_parser: StoreOperationParser) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        index = collection_name if collection_name is not None else self.index
        return index

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> PineconeCollection | None:
        if op_parser.is_resource_op():
            return None
        index_name = self._get_index_name(op_parser)
        if index_name is None:
            raise BadRequestError("Collection name must be specified")
        if index_name in self._collection_cache:
            return self._collection_cache[index_name]
        client = self._database_client.Index(index_name)
        description = self._database_client.describe_index(index_name)
        col = PineconeCollection(
            client,
            self.namespace,
            description.dimension,
            self.content_metadata_field,
        )
        self._collection_cache[index_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._database_client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: PineconeCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "index": self._get_index_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "index": self._get_index_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.fetch, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = converter.convert_put(
                op_parser.get_id_as_str(),
                op_parser.get_vector_value(),
                op_parser.get_metadata(),
            )
            call = NCall(client.upsert, args, nargs)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            args = converter.convert_update_value(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_updated_value(),
            )
            call = NCall(client.update, args, nargs)
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            args = converter.convert_update_metadata(
                id=op_parser.get_id_as_str(),
                metadata=op_parser.get_updated_metadata(),
            )
            call = NCall(client.update, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_id_as_str_or_none(), op_parser.get_where()
            )
            call = NCall(
                client.delete, args, nargs, {NotFoundException: NotFoundError}
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(client.query, args, nargs)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.describe_index_stats, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.fetch, args, nargs)
            elif func == "put":
                call = NCall(client.upsert, args, nargs)
            elif func == "delete":
                call = NCall(client.delete, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: PineconeCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = nresult.names()
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if len(nresult["vectors"]) == 0:
                raise NotFoundError
            for item in nresult["vectors"].values():
                result = self._convert_to_item(item, processor)
                break
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
                metadata=op_parser.get_updated_metadata(),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult["matches"]:
                items.append(self._convert_to_item(item, processor))
            items = QueryProcessor.order_items(
                items, op_parser.get_order_by(), processor.resolve_root_field
            )
            limit = op_parser.get_limit()
            items = QueryProcessor.limit_items(
                items,
                limit if limit is not None else DEFAULT_LIMIT,
                op_parser.get_offset(),
            )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            ns = self.namespace if self.namespace is not None else ""
            result = nresult["namespaces"][ns]["vector_count"]
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                for item in nresult["vectors"].values():
                    result.append(self._convert_to_item(item, processor))
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _convert_to_item(
        self, nitem: Any, processor: ItemProcessor
    ) -> VectorItem:
        content = None
        metadata: dict | None = None
        key = VectorKey(id=nitem["id"])
        if "metadata" in nitem:
            metadata = nitem["metadata"]
            if (
                metadata is not None
                and self.content_metadata_field is not None
                and self.content_metadata_field in metadata
            ):
                content = metadata[self.content_metadata_field]
                metadata.pop(self.content_metadata_field)
                if len(metadata) == 0:
                    metadata = None
        if "sparse_values" in nitem:
            sparse_values = nitem["sparse_values"]
            sparse_vector = dict()
            for i in range(0, len(sparse_values["indices"])):
                sparse_vector[sparse_values["indices"][i]] = sparse_values[
                    "values"
                ][i]
        else:
            sparse_vector = None
        value = VectorValue(
            vector=nitem["values"],
            sparse_vector=sparse_vector,
            content=content,
        )
        properties = None
        if "score" in nitem:
            properties = VectorProperties(score=nitem["score"])
        return VectorItem(
            key=key,
            value=value,
            metadata=metadata,
            properties=properties,
        )


class ResourceHelper:
    database_client: PineconeClient

    def __init__(self, database_client: PineconeClient):
        self.database_client = database_client

    def create_collection(
        self, index: str, config: VectorCollectionConfig | None, nargs: Any
    ):
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            dimension = vector_index.dimension
            metric = vector_index.metric.value
        else:
            dimension = 4
            metric = "dotproduct"
        if (
            config is not None
            and config.nconfig is not None
            and "spec" in config.nconfig
        ):
            spec = config.nconfig["spec"]
        else:
            spec = PodSpec(environment="gcp-starter", pod_type="s1.x1")
        args = {
            "name": index,
            "dimension": dimension,
            "metric": metric,
            "spec": spec,
        }
        return NCall(self.database_client.create_index, args, nargs).invoke()

    def drop_collection(self, index: str, nargs: Any):
        args = {"name": index}
        return NCall(self.database_client.delete_index, args, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        return NCall(self.database_client.list_indexes, None, nargs).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class OperationConverter:
    processor: ItemProcessor
    namespace: str | None
    dimension: int
    content_metadata_field: str | None

    def __init__(
        self,
        processor: ItemProcessor,
        namespace: str | None,
        dimension: int,
        content_metadata_field: str | None,
    ) -> None:
        self.processor = processor
        self.namespace = namespace
        self.dimension = dimension
        self.content_metadata_field = content_metadata_field

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"ids": ids, "namespace": self.namespace}

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        vectors = []
        for op_parser in op_parsers:
            vectors.append(
                self.convert_put_vector(
                    op_parser.get_id_as_str(),
                    op_parser.get_vector_value(),
                    op_parser.get_metadata(),
                )
            )
        return {
            "vectors": vectors,
            "namespace": self.namespace,
        }

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"ids": ids, "namespace": self.namespace}

    def convert_get(self, id: str) -> dict:
        return {"ids": [id], "namespace": self.namespace}

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        return {
            "vectors": [self.convert_put_vector(id, value, metadata)],
            "namespace": self.namespace,
        }

    def convert_put_vector(
        self, id: str, value: dict, metadata: dict | None
    ) -> dict:
        args: dict = {"id": id}
        content = None
        md = None
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["values"] = vector_value.vector
        if vector_value.sparse_vector is not None:
            args["sparse_values"] = self.convert_sparse_vector(
                vector_value.sparse_vector
            )
        content = vector_value.content
        if metadata is None:
            if content is not None and self.content_metadata_field is not None:
                md = {self.content_metadata_field: content}
        else:
            md = copy.deepcopy(metadata)
            if content is not None and self.content_metadata_field is not None:
                md[self.content_metadata_field] = content
        if md is not None:
            args["metadata"] = md
        return args

    def convert_update_value(self, id: str, value: dict) -> dict:
        args: dict = {"id": id, "namespace": self.namespace}
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["values"] = vector_value.vector
        if vector_value.sparse_vector is not None:
            args["sparse_values"] = self.convert_sparse_vector(
                vector_value.sparse_vector
            )
        if (
            vector_value.content is not None
            and self.content_metadata_field is not None
        ):
            args["set_metadata"] = {
                self.content_metadata_field: vector_value.content
            }
        return args

    def convert_update_metadata(self, id: str, metadata: dict | None) -> dict:
        return {
            "id": id,
            "set_metadata": metadata,
            "namespace": self.namespace,
        }

    def convert_delete(self, id: str | None, where: Expression) -> dict:
        if id is None and where is None:
            return {"delete_all": True, "namespace": self.namespace}
        if id is not None and where is None:
            return {"ids": [id], "namespace": self.namespace}
        if where is not None and id is None:
            return {
                "filter": self.convert_expr(where),
                "namespace": self.namespace,
            }
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> dict:
        args: dict = dict()
        search_args = self.convert_search(search)
        args = args | search_args
        select_args = self.convert_select(select)
        if select_args is not None:
            args = args | select_args
        if where is not None:
            args["filter"] = self.convert_expr(where)
        _offset = offset if offset is not None else 0
        _limit = limit if limit is not None else DEFAULT_LIMIT
        args["top_k"] = _limit + _offset
        return args

    def convert_select(self, select: Select | None) -> dict:
        include_value = False
        include_metadata = False
        if select is None or len(select.terms) == 0:
            include_value = True
            include_metadata = True
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    # include metadata since content is stored there
                    # filter it out when returning
                    # optimize later
                    include_value = True
                    include_metadata = True
                elif is_metadata_field(term.field):
                    include_metadata = True
        return {
            "include_values": include_value,
            "include_metadata": include_metadata,
        }

    def convert_search(self, search: Function | None) -> dict:
        if search is None:
            return {"vector": [0.0] * self.dimension}
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                return {
                    "vector": args.vector,
                    "sparse_vector": self.convert_sparse_vector(
                        args.sparse_vector
                    ),
                }

        raise BadRequestError("Search function not supported")

    def convert_sparse_vector(
        self, sparse_vector: dict[int, float] | None
    ) -> dict | None:
        if sparse_vector is None:
            return None
        indices = []
        values = []
        for k, v in sparse_vector.items():
            indices.append(k)
            values.append(v)
        return {"indices": indices, "values": values}

    def convert_expr(self, expr: Expression) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return field
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            comparion_map = {
                ComparisonOp.EQ: "$eq",
                ComparisonOp.NEQ: "$ne",
                ComparisonOp.GT: "$gt",
                ComparisonOp.GTE: "$gte",
                ComparisonOp.LT: "$lt",
                ComparisonOp.LTE: "$lte",
                ComparisonOp.IN: "$in",
                ComparisonOp.NIN: "$nin",
            }
            op = expr.op
            if isinstance(expr.lexpr, Field):
                field = self.convert_expr(expr.lexpr)
                value = self.convert_expr(expr.rexpr)
            elif isinstance(expr.rexpr, Field):
                field = self.convert_expr(expr.rexpr)
                value = self.convert_expr(expr.lexpr)
                op = Comparison.reverse_op(op)

            if op in comparion_map:
                pc_op = comparion_map[op]
                return {field: {pc_op: value}}
            elif op == ComparisonOp.BETWEEN:
                return {
                    "$and": [
                        {field: {"$gte": value[0]}},
                        {field: {"$lte": value[1]}},
                    ]
                }
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            return {
                "$and": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Or):
            return {
                "$or": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Not):
            raise BadRequestError("Not expression is not supported")
        return str(expr)

    def convert_func(self, expr: Function) -> dict:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_expr(args[0])
                return {field: {"$exists": True}}
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(args[0])
                return {field: {"$exists": False}}
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(args[0])
                return {field: {"$eq": args[1]}}
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(args[0])
                return {field: {"$in": args[1]}}
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        if where is None:
            return None
        args = {"filter": self.convert_expr(where)}
        return args


class PineconeCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(
        self,
        client: Any,
        namespace: str | None,
        dimension: int,
        content_metadata_field: str | None,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            namespace=namespace,
            dimension=dimension,
            content_metadata_field=content_metadata_field,
        )



"""
Secret Store on AWS Secrets Manager.
"""

__all__ = ["AWSSecretsManager"]

from typing import Any

import boto3

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import NotFoundError, PreconditionFailedError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)
from verse.ql import OrderBy, OrderByTerm

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)


class AWSSecretsManager(StoreProvider):
    region_name: str | None
    profile_name: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    nparams: dict[str, Any]

    _client: Any
    _processor: ItemProcessor

    def __init__(
        self,
        region_name: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            region_name:
                AWS region name.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key id.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            nparams:
                Native parameters to boto3 client.
        """
        self.region_name = region_name
        self.profile_name = profile_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.nparams = nparams

        self._client = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        session = None
        if self.profile_name is not None:
            session = boto3.session.Session(
                profile_name=self.profile_name,
                **self.nparams,
            )
        elif (
            self.aws_access_key_id is not None
            and self.aws_secret_access_key is not None
        ):
            session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                **self.nparams,
            )
        else:
            session = boto3.session.Session(**self.nparams)

        self._client = session.client(
            service_name="secretsmanager",
            region_name=self.region_name,
            **self.nparams,
        )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        ex = client.exceptions
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            args: dict[str, Any] = {"SecretId": id}
            if version is not None:
                args["VersionId"] = version
            call = NCall(
                client.get_secret_value,
                args,
                nargs,
                {
                    ex.ResourceNotFoundException: NotFoundError,
                    ex.InvalidRequestException: NotFoundError,
                },
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"SecretId": id}
            call = NCall(
                client.describe_secret,
                args,
                nargs,
                {ex.ResourceNotFoundException: NotFoundError},
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"SecretId": id, "IncludeDeprecated": True}
            call = NCall(
                client.list_secret_version_ids,
                args,
                nargs,
                {ex.ResourceNotFoundException: NotFoundError},
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata = op_parser.get_metadata()
            exists = op_parser.get_where_exists()
            args = {
                "Name": id,
                "SecretString": value,
                "Tags": Helper.convert_metadata_to_list(metadata),
            }
            if exists is False:
                call = NCall(
                    client.create_secret,
                    args,
                    nargs,
                    {ex.ResourceExistsException: PreconditionFailedError},
                )
            elif exists is None or exists is True:
                args = {
                    "id": id,
                    "value": value,
                    "metadata": metadata if metadata is not None else {},
                    "where_exists": exists,
                    "nargs": nargs,
                }
                call = NCall(helper.put, args)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            args = {
                "SecretId": id,
                "SecretString": value,
            }
            call = NCall(
                client.update_secret,
                args,
                nargs,
                {ex.ResourceNotFoundException: NotFoundError},
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            args = {
                "id": id,
                "metadata": metadata if metadata is not None else {},
                "nargs": nargs,
            }
            call = NCall(
                helper.update_metadata,
                args,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            args = {"id": id, "nargs": nargs}
            call = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            result = SecretItem(
                key=SecretKey(
                    id=nresult["Name"], version=nresult["VersionId"]
                ),
                value=nresult["SecretString"],
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            raise_on_deleted = True
            nflags = op_parser.get_nflags()
            if (
                nflags is not None
                and "raise_on_deleted" in nflags
                and nflags["raise_on_deleted"] is False
            ):
                raise_on_deleted = False
            if "DeletedDate" in nresult:
                if raise_on_deleted:
                    raise NotFoundError
            result = SecretItem(
                key=SecretKey(id=nresult["Name"]),
                metadata=Helper.convert_list_to_metadata(nresult["Tags"]),
                properties=SecretProperties(
                    created_time=nresult["CreatedDate"]
                ),
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            versions: list = []
            for item in nresult["Versions"]:
                versions.append(
                    SecretVersion(
                        version=item["VersionId"],
                        created_time=item["CreatedDate"],
                    )
                )
            versions.reverse()
            return SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = SecretItem(
                key=SecretKey(id=nresult["Name"], version=nresult["VersionId"])
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = SecretItem(
                key=SecretKey(id=nresult["Name"], version=nresult["VersionId"])
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            result = SecretItem(
                key=SecretKey(id=id),
                metadata=metadata if metadata is not None else {},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for response in nresult:
                for item in response["SecretList"]:
                    if "DeletedDate" in item:
                        continue
                    items.append(
                        SecretItem(
                            key=SecretKey(id=item["Name"]),
                            metadata=Helper.convert_list_to_metadata(
                                item["Tags"]
                            ),
                            properties=SecretProperties(
                                created_time=item["CreatedDate"]
                            ),
                        )
                    )
            order_by = (
                op_parser.get_order_by()
                if op_parser.get_order_by() is not None
                else OrderBy(terms=[OrderByTerm(field="$id")])
            )
            items = QueryProcessor.query_items(
                items=items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=order_by,
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_root_field,
            )
            result = SecretList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = []
            for response in nresult:
                for item in response["SecretList"]:
                    if "DeletedDate" in item:
                        continue
                    items.append(
                        SecretItem(
                            key=SecretKey(id=item["Name"]),
                            metadata=Helper.convert_list_to_metadata(
                                item["Tags"]
                            ),
                            properties=SecretProperties(
                                created_time=item["CreatedDate"]
                            ),
                        )
                    )
            result = QueryProcessor.count_items(
                items=items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_root_field,
            )
        return result


class Helper:
    @staticmethod
    def convert_metadata_to_list(metadata: dict | None) -> list[dict]:
        if metadata is None:
            return []
        metadata_list = []
        for key, value in metadata.items():
            metadata_list.append({"Key": key, "Value": value})
        return metadata_list

    @staticmethod
    def convert_list_to_metadata(metadata_list: list[Any]) -> dict:
        if not metadata_list:
            return {}
        metadata: dict = {}
        for item in metadata_list:
            metadata[item["Key"]] = item["Value"]
        return metadata


class ClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool | None,
        nargs: Any,
    ) -> Any:
        ex = self.client.exceptions
        if where_exists is None:
            (response, error) = NCall(
                self.client.create_secret,
                {
                    "Name": id,
                    "SecretString": value,
                    "Tags": Helper.convert_metadata_to_list(metadata),
                },
                nargs,
                {ex.ResourceExistsException: None},
            ).invoke(return_error=True)
            if not error:
                return response
        response = NCall(
            self.client.update_secret,
            {"SecretId": id, "SecretString": value},
            nargs,
            {
                ex.ResourceNotFoundException: (
                    PreconditionFailedError
                    if where_exists is True
                    else NotFoundError
                )
            },
        ).invoke()
        self.update_metadata(id, metadata, nargs)
        return response

    def update_metadata(
        self, id: str, metadata: dict | None, nargs: Any
    ) -> Any:
        ex = self.client.exceptions
        response = NCall(
            self.client.describe_secret,
            {"SecretId": id},
            nargs,
            {ex.ResourceNotFoundException: NotFoundError},
        ).invoke()
        prev_metadata = [o["Key"] for o in response["Tags"]]
        NCall(
            self.client.untag_resource,
            {"SecretId": id, "TagKeys": prev_metadata},
            nargs,
            {ex.ResourceNotFoundException: NotFoundError},
        ).invoke()
        if metadata is not None:
            NCall(
                self.client.tag_resource,
                {
                    "SecretId": id,
                    "Tags": Helper.convert_metadata_to_list(metadata),
                },
                nargs,
                {ex.ResourceNotFoundException: NotFoundError},
            ).invoke()

    def delete(self, id: str, nargs: Any) -> Any:
        response = NCall(
            self.client.describe_secret,
            {"SecretId": id},
            nargs,
            {self.client.exceptions.ResourceNotFoundException: NotFoundError},
        ).invoke()
        if "DeletedDate" in response:
            raise NotFoundError
        NCall(
            self.client.delete_secret,
            {"SecretId": id, "ForceDeleteWithoutRecovery": True},
            nargs,
        ).invoke()

    def query(self, nargs: Any):
        responses = []
        next_token = None
        while True:
            args = {
                "MaxResults": 100,
                "Filters": [],
                "IncludePlannedDeletion": False,
            }
            if next_token is not None:
                args["NextToken"] = next_token
            response = NCall(self.client.list_secrets, args, nargs).invoke()
            responses.append(response)
            if "NextToken" in response:
                next_token = response["NextToken"]
            else:
                break
        return responses

    def close(self, nargs: Any) -> Any:
        pass



"""
Object Store on Azure Blob Storage.
"""

from __future__ import annotations

__all__ = ["AzureBlobStorage"]

import binascii
from datetime import datetime, timedelta, timezone
from typing import IO, Any

from azure.core.exceptions import (
    HttpResponseError,
    ResourceModifiedError,
    ResourceNotFoundError,
)
from azure.storage.blob import (
    BlobSasPermissions,
    ContentSettings,
    StandardBlobTier,
    generate_blob_sas,
)

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    NotFoundError,
    NotModified,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    MatchCondition,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import (
    QueryArgs,
    get_collection_config,
    get_query_args,
    get_transfer_config,
)
from .._models import (
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectStoreClass,
    ObjectTransferConfig,
    ObjectVersion,
)


class AzureBlobStorage(StoreProvider):
    container: str | None
    connection_string: str | None
    account_url: str | None
    account_key: str | None
    sas_token: str | None
    credential_type: str | None
    tenant_id: str | None
    client_id: str | None
    client_secret: str | None
    certificate_path: str | None
    nparams: dict[str, Any]

    _credential: Any
    _acredential: Any
    _service_client: Any
    _aservice_client: Any

    _collection_cache: dict[str, AzureBlobStorageCollection]
    _acollection_cache: dict[str, AzureBlobStorageCollection]

    def __init__(
        self,
        container: str | None = None,
        connection_string: str | None = None,
        account_url: str | None = None,
        account_key: str | None = None,
        sas_token: str | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            container:
                Azure blob storage container
                mapped to object store collection.
            connection_string:
                Azure blob storage connection string.
            account_url:
                Azure blob storage account url.
            account_key:
                Azure blob storage account key.
            sas_token:
                SAS token.
            credential_type:
                Azure credential type.
                Credential types are default, client_secret, certificate,
                azure_cli, shared_token_cache, managed_identity.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Azure blob storage client.
        """
        self.container = container
        self.connection_string = connection_string
        self.account_url = account_url
        self.account_key = account_key
        self.sas_token = sas_token
        self.credential_type = credential_type
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.certificate_path = certificate_path
        self.nparams = nparams

        self._service_client = None
        self._aservice_client = None
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._service_client is not None:
            return

        from azure.storage.blob import BlobServiceClient

        from verse.internal.storage_common._azure import get_credential

        (
            self._credential,
            self._service_client,
        ) = self._init_credential_client(
            BlobServiceClient,
            get_credential,
        )

    async def ainit(self, context: Context | None = None) -> None:
        if self._aservice_client is not None:
            return

        from azure.storage.blob.aio import BlobServiceClient

        from verse.internal.storage_common._azure import aget_credential

        (
            self._acredential,
            self._aservice_client,
        ) = self._init_credential_client(
            BlobServiceClient,
            aget_credential,
        )

    def _init_credential_client(
        self,
        service_client,
        get_credential,
    ):
        credential = None
        if self.connection_string is not None:
            client = service_client.from_connection_string(
                self.connection_string,
                **self.nparams,
            )
        elif self.account_key is not None:
            client = service_client(
                account_url=self.account_url,
                credential=self.account_key,
                **self.nparams,
            )
        elif self.sas_token is not None:
            client = service_client(
                account_url=self.account_url,
                credential=self.account_key,
                **self.nparams,
            )
        else:
            credential = get_credential(
                credential_type=self.credential_type,
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
                certificate_path=self.certificate_path,
            )
            client = service_client(
                account_url=self.account_url,
                credential=credential,
                **self.nparams,
            )
        return credential, client

    def _get_container_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        container = (
            collection_name if collection_name is not None else self.container
        )
        return container

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> AzureBlobStorageCollection | None:
        if op_parser.is_resource_op():
            return None
        container_name = self._get_container_name(op_parser)
        if container_name is None:
            raise BadRequestError("Collection name must be specified")
        if container_name in self._collection_cache:
            return self._collection_cache[container_name]

        from azure.storage.blob import BlobClient

        client = self._service_client.get_container_client(container_name)
        col = AzureBlobStorageCollection(
            self._service_client,
            client,
            ClientHelper,
            container_name,
            BlobClient,
        )
        self._collection_cache[container_name] = col
        return col

    async def _aget_collection(
        self, op_parser: StoreOperationParser
    ) -> AzureBlobStorageCollection | None:
        if op_parser.is_resource_op():
            return None
        container_name = self._get_container_name(op_parser)
        if container_name is None:
            raise BadRequestError("Collection name must be specified")
        if container_name in self._acollection_cache:
            return self._acollection_cache[container_name]

        from azure.storage.blob.aio import BlobClient

        client = self._aservice_client.get_container_client(container_name)
        col = AzureBlobStorageCollection(
            self._aservice_client,
            client,
            AsyncClientHelper,
            container_name,
            BlobClient,
        )
        self._acollection_cache[container_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._service_client, self._credential),
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = await self._aget_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            AsyncResourceHelper(self._aservice_client, self._acredential),
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: AzureBlobStorageCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict]:
        if collection is not None:
            op_converter = collection.op_converter
            client = collection.client
            helper = collection.helper
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()

        def get_error_map(match_condition: MatchCondition):
            return {
                ResourceModifiedError: (
                    PreconditionFailedError
                    if match_condition.exists is True
                    or match_condition.exists is False
                    or match_condition.if_match
                    or match_condition.if_unmodified_since
                    else NotModified
                ),
                ResourceNotFoundError: NotFoundError,
            }

        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "container": self._get_container_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "container": self._get_container_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            match_condition = op_parser.get_match_condition()
            args, blob_client, state = op_converter.convert_put(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_value_as_bytes(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=match_condition,
                returning=op_parser.get_returning(),
                config=get_transfer_config(op_parser),
            )
            call = NCall(
                blob_client.upload_blob,
                args,
                nargs,
                get_error_map(match_condition),
            )
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            match_condition = op_parser.get_match_condition()
            args, blob_client = op_converter.convert_get(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                match_condition=match_condition,
                start=op_parser.get_start(),
                end=op_parser.get_end(),
                config=get_transfer_config(op_parser),
            )
            call = NCall(
                helper.get,
                {
                    "blob_client": blob_client,
                    "args": args,
                    "file": op_parser.get_file(),
                    "stream": op_parser.get_stream(),
                    "nargs": nargs,
                },
                None,
                get_error_map(match_condition),
            )
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            match_condition = op_parser.get_match_condition()
            args, blob_client = op_converter.convert_get_properties(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=match_condition,
            )
            call = NCall(
                blob_client.get_blob_properties,
                args,
                nargs,
                get_error_map(match_condition),
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            args = op_converter.convert_get_versions(
                id=op_parser.get_id_as_str()
            )
            call = NCall(
                helper.get_versions, {"args": args, "nargs": nargs}, None
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            match_condition = op_parser.get_match_condition()
            args = {
                "id": op_parser.get_id_as_str(),
                "version": op_parser.get_version(),
                "metadata": op_parser.get_updated_metadata(),
                "properties": op_parser.get_updated_properities(),
                "match_condition": match_condition,
                "nargs": nargs,
            }
            call = NCall(
                helper.update,
                args,
                nargs,
                get_error_map(match_condition),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            match_condition = op_parser.get_match_condition()
            args, func = op_converter.convert_delete(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=match_condition,
            )
            if func == "helper_delete":
                args["nargs"] = nargs
                call = NCall(
                    helper.delete,
                    args,
                    None,
                    get_error_map(match_condition)
                    | {ResourceNotFoundError: None},
                )
            elif func == "delete_blob":
                call = NCall(
                    client.delete_blob,
                    args,
                    nargs,
                    get_error_map(match_condition)
                    | {ResourceNotFoundError: None},
                )
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            match_condition = op_parser.get_match_condition()
            args = {
                "id": op_parser.get_id_as_str(),
                "source_id": op_parser.get_source_id_as_str(),
                "source_version": op_parser.get_source_version(),
                "source_collection": op_parser.get_source_collection(),
                "metadata": op_parser.get_metadata(),
                "properties": op_parser.get_properties(),
                "match_condition": match_condition,
                "nargs": nargs,
            }
            call = NCall(
                helper.copy,
                args,
                nargs,
                get_error_map(match_condition),
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            args = op_converter.convert_generate(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                method=op_parser.get_method(),
                expiry=op_parser.get_expiry(),
            )
            call = NCall(helper.generate, {"args": args, "nargs": nargs}, None)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, func = op_converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "delete":
                args["nargs"] = nargs
                call = NCall(helper.batch_delete, args, nargs)
            state["func"] = func
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: StoreOperationParser,
        collection: AzureBlobStorageCollection | None,
    ) -> Any:
        if collection is not None:
            result_converter = collection.result_converter
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [c.name for c in nresult]
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            if "open_file" in state:
                state["open_file"].close()
            result = result_converter.convert_put_result(
                op_parser.get_id_as_str(), nresult
            )
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            result = result_converter.convert_get_result(
                op_parser.get_id_as_str(),
                op_parser.get_file(),
                op_parser.get_stream(),
                nresult,
            )
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            result = result_converter.convert_get_properties_result(
                op_parser.get_id_as_str(), nresult
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            result = result_converter.convert_get_versions_result(
                op_parser.get_id_as_str(), nresult
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = result_converter.convert_update_result(
                op_parser.get_id_as_str(), nresult
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            result = result_converter.convert_copy_result(
                op_parser.get_id_as_str(), nresult
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            result = result_converter.convert_generate_result(
                op_parser.get_id_as_str(),
                op_parser.get_version(),
                nresult,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            result = result_converter.convert_query_result(
                nresult,
                state["query_args"],
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = result_converter.convert_count_result(
                nresult,
            )
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            func = state["func"]
            if func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result


class ResourceHelper:
    service_client: Any
    credential: Any

    def __init__(self, service_client: Any, credential: Any):
        self.service_client = service_client
        self.credential = credential

    def create_collection(
        self,
        container: str,
        config: ObjectCollectionConfig | None,
        nargs: Any,
    ):
        args = {
            "name": container,
        }
        if config is not None and config.acl is not None:
            if config.acl == "public-read":
                args["public_access"] = "blob"
        if config is not None and config.nconfig is not None:
            args = args | config.nconfig
        NCall(
            self.service_client.create_container,
            args,
            nargs,
        ).invoke()

    def drop_collection(self, container: str, nargs: Any):
        args = {
            "container": container,
        }
        NCall(
            self.service_client.delete_container,
            args,
            nargs,
        ).invoke()

    def list_collections(self, nargs) -> Any:
        response = NCall(
            self.service_client.list_containers, None, nargs
        ).invoke()
        return response

    def close(self, nargs: Any) -> Any:
        pass


class AsyncResourceHelper:
    service_client: Any
    credential: Any

    def __init__(self, service_client: Any, credential: Any):
        self.service_client = service_client
        self.credential = credential

    async def create_collection(
        self,
        container: str,
        config: ObjectCollectionConfig | None,
        nargs: Any,
    ):
        args = {
            "name": container,
        }
        if config is not None and config.acl is not None:
            if config.acl == "public-read":
                args["public_access"] = "blob"
        if config is not None and config.nconfig is not None:
            args = args | config.nconfig
        await NCall(
            self.service_client.create_container,
            args,
            nargs,
        ).ainvoke()

    async def drop_collection(self, container: str, nargs: Any):
        args = {
            "container": container,
        }
        await NCall(
            self.service_client.delete_container,
            args,
            nargs,
        ).ainvoke()

    async def list_collections(self, nargs) -> Any:
        nresult = []
        response = NCall(
            self.service_client.list_containers, None, nargs
        ).invoke()
        async for item in response:
            nresult.append(item)
        return nresult

    async def close(self, nargs: Any) -> Any:
        await NCall(
            self.service_client.close,
            None,
            nargs,
        ).ainvoke()
        if self.credential is not None:
            await NCall(
                self.credential.close,
                None,
                nargs,
            ).ainvoke()


class ClientHelper:
    service_client: Any
    client: Any
    container: str
    op_converter: OperationConverter

    def __init__(
        self,
        service_client: Any,
        client: Any,
        container: str,
        op_converter: OperationConverter,
    ):
        self.service_client = service_client
        self.client = client
        self.container = container
        self.op_converter = op_converter

    def get(
        self,
        blob_client: Any,
        args: dict,
        file: str | None,
        stream: IO | None,
        nargs: Any,
    ) -> Any:
        nresult = NCall(blob_client.download_blob, args, nargs).invoke()
        value = None
        if file is not None:
            with open(file=file, mode="wb") as f:
                f.write(nresult.readall())
        elif stream is not None:
            nresult.readinto(stream)
        if file is None and stream is None:
            value = nresult.readall()
        return {"blob": nresult, "value": value}

    def get_versions(self, args: dict, nargs: Any):
        nresult = NCall(self.client.list_blobs, args, nargs).invoke()
        return [blob for blob in nresult]

    def generate(self, args: dict, nargs: Any) -> Any:
        return NCall(generate_blob_sas, args, nargs).invoke()

    def delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
        nargs: Any,
    ) -> Any:
        args: dict = {"blob": id}
        args = args | self.op_converter.convert_match_condition(
            match_condition
        )
        NCall(self.client.delete_blob, args, nargs).invoke()
        args = {"name_starts_with": id, "include": ["versions"]}
        blobs = NCall(self.client.list_blobs, args, None).invoke()
        blob_properties = []
        for blob in blobs:
            if blob.name == id:
                blob_properties.append(blob)
        NCall(self.client.delete_blobs, blob_properties, nargs).invoke()
        return None

    def copy(
        self,
        id: str,
        source_id: str,
        source_version: str | None,
        source_collection: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        nargs: Any,
    ):
        sas = NCall(
            generate_blob_sas,
            {
                "account_name": self.service_client.account_name,
                "container_name": (
                    source_collection
                    if source_collection is not None
                    else self.container
                ),
                "blob_name": source_id,
                "account_key": self.service_client.credential.account_key,
                "version_id": source_version,
                "permission": BlobSasPermissions(read=True),
                "expiry": datetime.now(timezone.utc) + timedelta(seconds=3600),
            },
            None,
        ).invoke()
        source_blob_client = self.service_client.get_blob_client(
            container=(
                source_collection
                if source_collection is not None
                else self.container
            ),
            blob=source_id,
        )
        source_url = f"{source_blob_client.url}?{sas}"
        args, blob_client, state = self.op_converter.convert_put(
            id,
            value=b"",
            file=None,
            stream=None,
            metadata=metadata,
            properties=properties,
            match_condition=match_condition,
            returning=None,
            config=None,
            copy=True,
        )
        args.pop("data")
        args["source_url"] = source_url
        args["overwrite"] = True
        try:
            return NCall(
                blob_client.upload_blob_from_url, args, nargs
            ).invoke()
        except HttpResponseError as e:
            if e.status_code == 404:
                raise NotFoundError
            if e.status_code == 412:
                raise PreconditionFailedError
            raise

    def update(
        self,
        id: str,
        version: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        nargs: Any,
    ):
        put_args = self.op_converter.convert_properties_metadata(
            properties, metadata
        )
        match_args = self.op_converter.convert_match_condition(match_condition)
        blob_client = self.service_client.get_blob_client(
            container=self.container,
            blob=id,
        )
        response = None
        if "standard_blob_tier" in put_args:
            args = {"standard_blob_tier": put_args["standard_blob_tier"]}
            args = args | match_args
            if version is not None:
                args["version_id"] = version
            response = NCall(
                blob_client.set_standard_blob_tier,
                args,
                nargs,
            ).invoke()
        if "content_settings" in put_args:
            args = {"content_settings": put_args["content_settings"]}
            args = args | match_args
            response = NCall(
                blob_client.set_http_headers,
                args,
                nargs,
            ).invoke()
        if "metadata" in put_args:
            if response is not None and "if_match" in match_args:
                match_args["if_match"] = response["etag"]
            args = {"metadata": put_args["metadata"]}
            args = args | match_args
            response = NCall(
                blob_client.set_blob_metadata, args, nargs
            ).invoke()
        return response

    def query(
        self,
        query_args: QueryArgs,
        nargs: Any,
    ) -> Any:
        from azure.storage.blob import BlobPrefix

        blobs: list = []
        prefixes: list = []
        args: dict = {}
        if query_args.prefix is not None:
            args["name_starts_with"] = query_args.prefix
        if query_args.delimiter is not None:
            args["delimiter"] = query_args.delimiter
        continuation_token = query_args.continuation
        total_key_count = 0
        if query_args.limit is not None and query_args.page_size is not None:
            if query_args.page_size < query_args.limit - total_key_count:
                args["results_per_page"] = query_args.page_size
            else:
                args["results_per_page"] = query_args.limit - total_key_count
        elif query_args.limit is not None:
            args["results_per_page"] = query_args.limit - total_key_count
        elif query_args.page_size is not None:
            args["results_per_page"] = query_args.page_size
        if query_args.delimiter is not None:
            iterator = NCall(self.client.walk_blobs, args, nargs).invoke()
        else:
            iterator = NCall(self.client.list_blobs, args, nargs).invoke()
        iterator_args = {}
        if continuation_token is not None:
            iterator_args["continuation_token"] = continuation_token
        blob_pager = NCall(iterator.by_page, iterator_args, None).invoke()
        for page in blob_pager:
            for b in page:
                if isinstance(b, BlobPrefix):
                    prefixes.append(b.name)
                else:
                    blobs.append(b)
                total_key_count = total_key_count + 1
                if (
                    query_args.limit is not None
                    and total_key_count >= query_args.limit
                ):
                    break
            if (
                query_args.limit is not None
                and total_key_count >= query_args.limit
            ):
                break
            if blob_pager.continuation_token is not None:
                continuation_token = blob_pager.continuation_token
            else:
                continuation_token = None
                break
            if query_args.paging:
                break
        return dict(
            blobs=blobs, prefixes=prefixes, continuation=continuation_token
        )

    def batch_delete(self, blobs: list, nargs: Any) -> None:
        for blob in blobs:
            NCall(
                self.client.delete_blob,
                {"blob": blob["id"], "version_id": blob["version"]},
                nargs,
            ).invoke()


class AsyncClientHelper:
    service_client: Any
    client: Any
    container: str
    op_converter: OperationConverter

    def __init__(
        self,
        service_client: Any,
        client: Any,
        container: str,
        op_converter: OperationConverter,
    ):
        self.service_client = service_client
        self.client = client
        self.container = container
        self.op_converter = op_converter

    async def get(
        self,
        blob_client: Any,
        args: dict,
        file: str | None,
        stream: IO | None,
        nargs: Any,
    ) -> Any:
        nresult = await NCall(blob_client.download_blob, args, nargs).ainvoke()
        value = None
        if file is not None:
            with open(file=file, mode="wb") as f:
                f.write(await nresult.readall())
        elif stream is not None:
            await nresult.readinto(stream)
        if file is None and stream is None:
            value = await nresult.readall()
        return {"blob": nresult, "value": value}

    async def get_versions(self, args: dict, nargs: Any):
        nresult = NCall(self.client.list_blobs, args, nargs).invoke()
        return [blob async for blob in nresult]

    async def generate(self, args: dict, nargs: Any) -> Any:
        return NCall(generate_blob_sas, args, nargs).invoke()

    async def delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
        nargs: Any,
    ) -> Any:
        args: dict = {"blob": id}
        args = args | self.op_converter.convert_match_condition(
            match_condition
        )
        await NCall(self.client.delete_blob, args, nargs).ainvoke()
        args = {"name_starts_with": id, "include": ["versions"]}
        blobs = NCall(self.client.list_blobs, args, None).invoke()
        blob_properties = []
        async for blob in blobs:
            if blob.name == id:
                blob_properties.append(blob)
        await NCall(self.client.delete_blobs, blob_properties, nargs).ainvoke()
        return None

    async def copy(
        self,
        id: str,
        source_id: str,
        source_version: str | None,
        source_collection: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        nargs: Any,
    ):
        sas = NCall(
            generate_blob_sas,
            {
                "account_name": self.service_client.account_name,
                "container_name": (
                    source_collection
                    if source_collection is not None
                    else self.container
                ),
                "blob_name": source_id,
                "account_key": self.service_client.credential.account_key,
                "version_id": source_version,
                "permission": BlobSasPermissions(read=True),
                "expiry": datetime.now(timezone.utc) + timedelta(seconds=3600),
            },
            None,
        ).invoke()
        source_blob_client = self.service_client.get_blob_client(
            container=(
                source_collection
                if source_collection is not None
                else self.container
            ),
            blob=source_id,
        )
        source_url = f"{source_blob_client.url}?{sas}"
        args, blob_client, state = self.op_converter.convert_put(
            id,
            value=b"",
            file=None,
            stream=None,
            metadata=metadata,
            properties=properties,
            match_condition=match_condition,
            returning=None,
            config=None,
            copy=True,
        )
        args.pop("data")
        args["source_url"] = source_url
        args["overwrite"] = True
        try:
            return await NCall(
                blob_client.upload_blob_from_url, args, nargs
            ).ainvoke()
        except HttpResponseError as e:
            if e.status_code == 404:
                raise NotFoundError
            if e.status_code == 412:
                raise PreconditionFailedError
            raise

    async def update(
        self,
        id: str,
        version: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        nargs: Any,
    ):
        put_args = self.op_converter.convert_properties_metadata(
            properties, metadata
        )
        match_args = self.op_converter.convert_match_condition(match_condition)
        blob_client = self.service_client.get_blob_client(
            container=self.container,
            blob=id,
        )
        response = None
        if "standard_blob_tier" in put_args:
            args = {"standard_blob_tier": put_args["standard_blob_tier"]}
            args = args | match_args
            if version is not None:
                args["version_id"] = version
            response = await NCall(
                blob_client.set_standard_blob_tier,
                args,
                nargs,
            ).ainvoke()
        if "content_settings" in put_args:
            args = {"content_settings": put_args["content_settings"]}
            args = args | match_args
            response = await NCall(
                blob_client.set_http_headers,
                args,
                nargs,
            ).ainvoke()
        if "metadata" in put_args:
            if response is not None and "if_match" in match_args:
                match_args["if_match"] = response["etag"]
            args = {"metadata": put_args["metadata"]}
            args = args | match_args
            response = await NCall(
                blob_client.set_blob_metadata, args, nargs
            ).ainvoke()
        return response

    async def query(
        self,
        query_args: QueryArgs,
        nargs: Any,
    ) -> Any:
        from azure.storage.blob.aio import BlobPrefix

        blobs: list = []
        prefixes: list = []
        args: dict = {}
        if query_args.prefix is not None:
            args["name_starts_with"] = query_args.prefix
        if query_args.delimiter is not None:
            args["delimiter"] = query_args.delimiter
        continuation_token = query_args.continuation
        total_key_count = 0
        if query_args.limit is not None and query_args.page_size is not None:
            if query_args.page_size < query_args.limit - total_key_count:
                args["results_per_page"] = query_args.page_size
            else:
                args["results_per_page"] = query_args.limit - total_key_count
        elif query_args.limit is not None:
            args["results_per_page"] = query_args.limit - total_key_count
        elif query_args.page_size is not None:
            args["results_per_page"] = query_args.page_size
        if query_args.delimiter is not None:
            iterator = NCall(self.client.walk_blobs, args, nargs).invoke()
        else:
            iterator = NCall(self.client.list_blobs, args, nargs).invoke()
        iterator_args = {}
        if continuation_token is not None:
            iterator_args["continuation_token"] = continuation_token
        blob_pager = NCall(iterator.by_page, iterator_args, None).invoke()

        async for page in blob_pager:
            async for b in page:
                if isinstance(b, BlobPrefix):
                    prefixes.append(b.name)
                else:
                    blobs.append(b)
                total_key_count = total_key_count + 1
                if (
                    query_args.limit is not None
                    and total_key_count >= query_args.limit
                ):
                    break
            if (
                query_args.limit is not None
                and total_key_count >= query_args.limit
            ):
                break
            if blob_pager.continuation_token is not None:
                continuation_token = blob_pager.continuation_token
            else:
                continuation_token = None
                break
            if query_args.paging:
                break
        return dict(
            blobs=blobs, prefixes=prefixes, continuation=continuation_token
        )

    async def batch_delete(self, blobs: list, nargs: Any) -> None:
        import asyncio

        tasks = []
        for blob in blobs:
            tasks.append(
                NCall(
                    self.client.delete_blob,
                    {"blob": blob["id"], "version_id": blob["version"]},
                    nargs,
                ).ainvoke()
            )
        await asyncio.gather(*tasks)


class ResultConverter:
    service_client: Any
    processor: ItemProcessor
    container: str

    def __init__(
        self, service_client: Any, processor: ItemProcessor, container: str
    ):
        self.service_client = service_client
        self.processor = processor
        self.container = container

    def convert_md5(self, md5: bytearray | None) -> str | None:
        if md5 is None:
            return None
        blobmd5 = bytearray(md5)
        hex = binascii.hexlify(blobmd5).decode("utf-8")
        return hex

    def convert_url(self, id: str):
        return f"{self.service_client.url}{self.container}/{id}"

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            "Hot": ObjectStoreClass.HOT,
            "Cool": ObjectStoreClass.COOL,
            "Cold": ObjectStoreClass.COLD,
            "Archive": ObjectStoreClass.ARCHIVE,
        }
        if storage_class is None:
            return ObjectStoreClass.HOT.value
        if storage_class in map:
            return map[storage_class].value
        return storage_class

    def convert_properties(self, nresult: Any) -> ObjectProperties:
        properties = ObjectProperties()
        properties.cache_control = nresult.content_settings.cache_control
        properties.content_disposition = (
            nresult.content_settings.content_disposition
        )
        properties.content_encoding = nresult.content_settings.content_encoding
        properties.content_language = nresult.content_settings.content_language
        properties.content_md5 = self.convert_md5(
            nresult.content_settings.content_md5
        )
        properties.content_type = nresult.content_settings.content_type
        properties.content_length = nresult.size
        properties.last_modified = nresult.last_modified
        properties.etag = nresult.etag
        properties.storage_class = self.convert_storage_class(
            nresult.blob_tier
        )
        return properties

    def convert_put_result(self, id: str, nresult: Any) -> ObjectItem:
        version = None
        etag = None
        last_modified = None
        content_md5 = None
        if nresult is not None and isinstance(nresult, dict):
            version = nresult.get("version_id", None)
            etag = nresult.get("etag", None)
            last_modified = nresult.get("last_modified", None)
            content_md5 = nresult.get("content_md5", None)
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            properties=ObjectProperties(
                etag=etag,
                last_modified=last_modified,
                content_md5=self.convert_md5(content_md5),
            ),
            url=self.convert_url(id=id),
        )

    def convert_copy_result(self, id: str, nresult: Any) -> ObjectItem:
        version = None
        etag = None
        last_modified = None
        content_md5 = None
        metadata = None
        if nresult is not None and isinstance(nresult, dict):
            version = nresult.get("version_id", None)
            etag = nresult.get("etag", None)
            last_modified = nresult.get("last_modified", None)
            content_md5 = nresult.get("content_md5", None)
            metadata = nresult.get("metadata", None)
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            metadata=metadata,
            properties=ObjectProperties(
                etag=etag,
                last_modified=last_modified,
                content_md5=self.convert_md5(content_md5),
            ),
            url=self.convert_url(id=id),
        )

    def convert_get_result(
        self,
        id: str,
        file: str | None,
        stream: IO | None,
        nresult: Any,
    ) -> ObjectItem:
        blob = nresult["blob"]
        value = nresult["value"]
        return ObjectItem(
            key=ObjectKey(id=id, version=blob.properties.version_id),
            value=value,
            metadata=blob.properties.metadata,
            properties=self.convert_properties(blob.properties),
            url=self.convert_url(id),
        )

    def convert_get_properties_result(
        self, id: str, nresult: Any
    ) -> ObjectItem:
        key = ObjectKey(id=id, version=nresult.version_id)
        properties = self.convert_properties(nresult)
        return ObjectItem(
            key=key,
            metadata=nresult.metadata,
            properties=properties,
            url=self.convert_url(id),
        )

    def convert_get_versions_result(self, id: str, nresult: Any) -> ObjectItem:
        key = ObjectKey(id=id)
        versions: list = []
        url = None
        for blob in nresult:
            if blob.name != id or blob.deleted:
                continue
            properties = self.convert_properties(blob)
            versions.append(
                ObjectVersion(
                    version=blob.version_id,
                    properties=properties,
                    metadata=blob.metadata,
                    latest=blob.is_current_version,
                )
            )
            url = self.convert_url(id)
        if len(versions) == 0:
            raise NotFoundError
        return ObjectItem(key=key, versions=versions, url=url)

    def convert_update_result(self, id: str, nresult: Any) -> ObjectItem:
        version = None
        etag = None
        last_modified = None
        if "etag" in nresult:
            etag = nresult["etag"]
        if "last_modified" in nresult:
            last_modified = nresult["last_modified"]
        if "version_id" in nresult:
            version = nresult["version_id"]
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            properties=ObjectProperties(
                etag=etag, last_modified=last_modified
            ),
            url=self.convert_url(id),
        )

    def convert_generate_result(
        self, id: str, version: str | None, nresult: Any
    ) -> ObjectItem:
        blob_client = self.service_client.get_blob_client(
            container=self.container, blob=id
        )
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            url=f"{blob_client.url}?{nresult}",
        )

    def convert_query_result(
        self, nresult: Any, query_args: QueryArgs
    ) -> ObjectList:
        blobs = nresult["blobs"]
        prefixes = nresult["prefixes"]
        continuation = nresult["continuation"]
        items: list = []
        for blob in blobs:
            items.append(
                ObjectItem(
                    key=ObjectKey(id=blob.name),
                    properties=self.convert_properties(blob),
                    metadata=blob.metadata,
                    url=self.convert_url(blob.name),
                )
            )
        return ObjectList(
            items=items,
            continuation=continuation,
            prefixes=prefixes,
        )

    def convert_count_result(
        self,
        nresult: Any,
    ) -> int:
        blobs = nresult["blobs"]
        prefixes = nresult["prefixes"]
        return len(blobs) + len(prefixes)


class OperationConverter:
    service_client: Any
    processor: ItemProcessor
    container: str
    blob_client_type: Any

    def __init__(
        self,
        service_client: Any,
        processor: ItemProcessor,
        container: str,
        blob_client_type: Any,
    ):
        self.service_client = service_client
        self.processor = processor
        self.container = container
        self.blob_client_type = blob_client_type

    def convert_md5(self, md5_str: str | None) -> bytearray | None:
        if md5_str is None:
            return None
        blobmd5 = binascii.unhexlify(md5_str)
        return bytearray(blobmd5)

    def convert_transfer_config_upload_for_client(
        self, config: ObjectTransferConfig | None
    ) -> dict:
        args: dict = {}
        if config is not None:
            if config.multipart is not None:
                args["max_single_put_size"] = 0
            if config.chunksize is not None:
                args["max_block_size"] = config.chunksize
            if config.nconfig is not None:
                args = args | config.nconfig
        return args

    def convert_transfer_config_download_for_client(
        self, config: ObjectTransferConfig | None
    ) -> dict:
        args: dict = {}
        if config is not None:
            if config.multipart is not None:
                args["max_single_get_size"] = 1
            if config.chunksize is not None:
                args["max_chunk_get_size"] = config.chunksize
            if config.nconfig is not None:
                args = args | config.nconfig
        return args

    def convert_transfer_config(
        self, config: ObjectTransferConfig | None
    ) -> dict:
        args: dict = {}
        if config is not None:
            if config.concurrency is not None:
                args["max_concurrency"] = config.concurrency
            if config.nconfig is not None:
                args = args | config.nconfig
        return args

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            ObjectStoreClass.HOT: StandardBlobTier.HOT,
            ObjectStoreClass.COOL: StandardBlobTier.COOL,
            ObjectStoreClass.COLD: StandardBlobTier.COLD,
            ObjectStoreClass.ARCHIVE: StandardBlobTier.ARCHIVE,
        }
        if storage_class is None:
            return map[ObjectStoreClass.HOT]
        if ObjectStoreClass(storage_class) in map:
            return map[ObjectStoreClass(storage_class)]
        return storage_class

    def convert_properties_metadata(
        self, properties: dict | None, metadata: dict | None
    ) -> dict:
        args: dict = {}
        if properties is None:
            props = ObjectProperties()
        else:
            props = ObjectProperties(**properties)
        if (
            props.cache_control is not None
            or props.content_disposition is not None
            or props.content_encoding is not None
            or props.content_language is not None
            or props.content_md5 is not None
            or props.content_type is not None
        ):
            content_settings = ContentSettings()
            if props.cache_control is not None:
                content_settings.cache_control = props.cache_control
            if props.content_disposition is not None:
                content_settings.content_disposition = (
                    props.content_disposition
                )
            if props.content_encoding is not None:
                content_settings.content_encoding = props.content_encoding
            if props.content_language is not None:
                content_settings.content_language = props.content_language
            if props.content_md5 is not None:
                content_settings.content_md5 = self.convert_md5(
                    props.content_md5
                )
            if props.content_type is not None:
                content_settings.content_type = props.content_type
            args["content_settings"] = content_settings
        if props.storage_class is not None:
            args["standard_blob_tier"] = self.convert_storage_class(
                props.storage_class
            )
        if metadata is not None:
            args["metadata"] = metadata
        return args

    def convert_match_condition(self, match_condition: MatchCondition) -> dict:
        args: dict = {}
        if match_condition.exists is True:
            args["if_match"] = "*"
        if match_condition.exists is False:
            args["overwrite"] = False
        if match_condition.if_match is not None:
            args["if_match"] = match_condition.if_match
        if match_condition.if_none_match is not None:
            args["if_none_match"] = match_condition.if_none_match
        if match_condition.if_modified_since is not None:
            args["if_modified_since"] = match_condition.if_modified_since
        if match_condition.if_unmodified_since is not None:
            args["if_unmodified_since"] = match_condition.if_unmodified_since
        return args

    def convert_copy_match_condition(
        self, match_condition: MatchCondition
    ) -> dict:
        args: dict = {}
        if match_condition.exists is True:
            args["source_if_match"] = "*"
        if match_condition.if_match is not None:
            args["source_if_match"] = match_condition.if_match
        if match_condition.if_none_match is not None:
            args["source_if_none_match"] = match_condition.if_none_match
        if match_condition.if_modified_since is not None:
            args["source_if_modified_since"] = (
                match_condition.if_modified_since
            )
        if match_condition.if_unmodified_since is not None:
            args["source_if_unmodified_since"] = (
                match_condition.if_unmodified_since
            )
        return args

    def convert_put(
        self,
        id: str,
        value: bytes | None,
        file: str | None,
        stream: IO | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        returning: bool | None,
        config: ObjectTransferConfig | None,
        copy: bool = False,
    ) -> tuple[dict, Any, dict]:
        if config is not None and config.multipart:
            bc_args = {
                "account_url": self.service_client.url,
                "container_name": self.container,
                "blob_name": id,
                "credential": self.service_client.credential,
            }
            bc_args = bc_args | self.convert_transfer_config_upload_for_client(
                config
            )
            blob_client = self.blob_client_type(**bc_args)
        else:
            blob_client = self.service_client.get_blob_client(
                container=self.container, blob=id
            )

        args: dict = {}
        state: dict = {}
        if value is not None:
            args["data"] = value
        elif stream is not None:
            args["data"] = stream
        elif file is not None:
            args["data"] = open(file, "rb")
            state["open_file"] = args["data"]
        args = args | self.convert_properties_metadata(properties, metadata)
        if not copy:
            match_condition_args = self.convert_match_condition(
                match_condition
            )
            args = args | match_condition_args
            if (
                match_condition.exists is None
                and len(match_condition_args) == 0
            ):
                args["overwrite"] = True
        else:
            args = args | self.convert_copy_match_condition(match_condition)
        args = args | self.convert_transfer_config(config)
        return args, blob_client, state

    def convert_get(
        self,
        id: str,
        version: str | None,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        start: int | None,
        end: int | None,
        config: ObjectTransferConfig | None,
    ) -> tuple[dict, Any]:
        if config is not None and config.multipart:
            bc_args = {
                "account_url": self.service_client.url,
                "container_name": self.container,
                "blob_name": id,
                "credential": self.service_client.credential,
            }
            bc_args = (
                bc_args
                | self.convert_transfer_config_download_for_client(config)
            )
            blob_client = self.blob_client_type(**bc_args)
        else:
            blob_client = self.service_client.get_blob_client(
                container=self.container, blob=id
            )

        args: dict = {}
        if start is not None:
            args["offset"] = start
        if end is not None:
            s = start if start is not None else 0
            args["offset"] = s
            args["length"] = end - s + 1
        if version is not None:
            args["version_id"] = version
        args = args | self.convert_match_condition(match_condition)
        args = args | self.convert_transfer_config(config)
        return args, blob_client

    def convert_get_properties(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
    ) -> tuple[dict, Any]:
        args = {}
        blob_client = self.service_client.get_blob_client(
            container=self.container, blob=id
        )
        if version is not None:
            args["version_id"] = version
        args = args | self.convert_match_condition(match_condition)
        return args, blob_client

    def convert_get_versions(self, id: str) -> dict:
        return {"name_starts_with": id, "include": ["versions"]}

    def convert_delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
    ) -> tuple[dict, str]:
        if version == "*":
            args = {
                "id": id,
                "version": version,
                "match_condition": match_condition,
            }
            return args, "helper_delete"
        args = {"blob": id}
        if version is not None:
            args["version_id"] = version
        args = args | self.convert_match_condition(match_condition)
        return args, "delete_blob"

    def convert_generate(
        self,
        id: str,
        version: str | None,
        method: str | None,
        expiry: int | None,
    ):
        args = {}
        args["account_name"] = self.service_client.account_name
        args["container_name"] = self.container
        args["blob_name"] = id
        args["account_key"] = self.service_client.credential.account_key
        if method is None or method == "GET":
            args["permission"] = BlobSasPermissions(read=True)
        elif method == "PUT":
            args["permission"] = BlobSasPermissions(write=True)
        if expiry is not None:
            args["expiry"] = datetime.now(timezone.utc) + timedelta(
                seconds=expiry
            )
        if version is not None:
            args["version_id"] = version
        return args

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        for op_parser in op_parsers:
            if not op_parser.op_equals(StoreOperation.DELETE):
                raise BadRequestError("Operation not supported in BATCH")
        return self.convert_batch_delete(op_parsers), "delete"

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        blobs = []
        for op_parser in op_parsers:
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            blobs.append(dict(id=id, version=version))
        return {"blobs": blobs}


class AzureBlobStorageCollection:
    client: Any
    helper: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    processor: ItemProcessor

    def __init__(
        self,
        service_client: Any,
        client: Any,
        helper_type: Any,
        container: str,
        blob_client_type: Any,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.op_converter = OperationConverter(
            service_client,
            self.processor,
            container,
            blob_client_type,
        )
        self.result_converter = ResultConverter(
            service_client, self.processor, container
        )
        self.helper = helper_type(
            service_client, client, container, self.op_converter
        )



from enum import Enum
from typing import Any

from verse.core import DataModel


class VectorSearchArgs(DataModel):
    """Vector search args."""

    vector: list[float] | None = None
    """Vector floats."""

    sparse_vector: dict[int, float] | None = None
    """Sparse vector dictionary."""

    field: str | None = None
    """Vector field name."""


class VectorValue(DataModel):
    """Vector value."""

    vector: list[float] | None = None
    """Vector floats."""

    sparse_vector: dict[int, float] | None = None
    """Sparse vector dictionary."""

    content: str | None = None
    """Text context."""


class VectorKey(DataModel):
    """Vector key."""

    id: str
    """Vector id."""


class VectorProperties(DataModel):
    """Vector properties."""

    score: float | None = None
    """Match score."""


class VectorItem(DataModel):
    """Vector item."""

    key: VectorKey
    """Vector key."""

    value: VectorValue | None = None
    """Vector value."""

    metadata: dict[str, Any] | None = None
    """Vector metadata."""

    properties: VectorProperties | None = None
    """Vector properties."""


class VectorList(DataModel):
    """Vector list."""

    items: list[VectorItem]
    """List of vector items."""


class VectorIndexMetric(str, Enum):
    """Vector index metric."""

    DOTPRODUCT = "dotproduct"
    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    MANHATTAN = "manhattan"
    HAMMING = "hamming"


class VectorIndexStructure(str, Enum):
    """Vector index structure."""

    FLAT = "flat"
    HNSW = "hnsw"


class VectorIndex(DataModel):
    """Vector index."""

    field: str | None = None
    """Index field name."""

    dimension: int = 4
    """Vector dimension."""

    metric: VectorIndexMetric = VectorIndexMetric.DOTPRODUCT
    """Index metric."""

    structure: VectorIndexStructure | None = None
    """Vector index structure."""

    nconfig: dict | None = None
    """Native config."""


class SparseVectorIndex(DataModel):
    """Sparse vector index."""

    field: str | None = None
    """Sparse vector field name."""

    nconfig: dict | None = None
    """Native config."""


class VectorCollectionConfig(DataModel):
    """Vector collection config."""

    vector_index: VectorIndex | None = None
    """Vector index."""

    sparse_vector_index: SparseVectorIndex | None = None
    """Sparse vector index."""

    nconfig: dict[str, Any] | None = None
    """Native config."""



"""
In Memory Document Store.
"""

from __future__ import annotations

__all__ = ["Memory"]

from threading import Lock
from typing import Any

from verse.core import Context, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    ParameterParser,
    QueryProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)

from .._helper import build_item_from_value, build_query_result


class Memory(StoreProvider):
    collection: str | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None
    nparams: dict[str, Any]

    # (collection, key, value)
    _db: dict[str, dict[str, Any]]
    _collection_cache: dict[str, MemoryCollection]
    _lock: Lock

    def __init__(
        self,
        collection: str | None = None,
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            collection:
                Collection name.
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native parameters to the client. Not used.
        """
        self.collection = collection
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._db = dict()
        self._collection_cache = dict()
        self._lock = Lock()

    def init(self, context: Context | None = None) -> None:
        pass

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        db_collection = (
            collection_name if collection_name is not None else self.collection
        )
        return db_collection

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[MemoryCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[MemoryCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        db_collection = self._get_collection_name(op_parser)
        if db_collection is None:
            raise BadRequestError("Collection name must be specified")
        if db_collection in self._collection_cache:
            return [self._collection_cache[db_collection]]
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field, db_collection
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field, db_collection
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, db_collection
        )
        if db_collection not in self._db:
            self._db[db_collection] = dict()
        col = MemoryCollection(
            self._db[db_collection],
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[db_collection] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        if len(collections) == 1:
            processor = collections[0].processor
            data = collections[0].data
        result: Any = None

        def get_db_key_from_key(key):
            return tuple(
                sorted(processor.get_normalized_key_from_key(key).items())
            )

        def get_db_key_from_value(value):
            return tuple(
                sorted(
                    processor.get_normalized_key_from_value(document).items()
                )
            )

        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            collection_name = self._get_collection_name(op_parser)
            with self._lock:
                if (
                    collection_name is not None
                    and collection_name not in self._db
                ):
                    self._db[collection_name] = dict()
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            collection_name = self._get_collection_name(op_parser)
            with self._lock:
                if collection_name in self._db:
                    self._db.pop(collection_name)
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = list(self._db.keys())
        # CREATE INDEX
        elif op_parser.op_equals(StoreOperation.CREATE_INDEX):
            result = None
        # DROP INDEX
        elif op_parser.op_equals(StoreOperation.DROP_INDEX):
            result = None
        # LIST INDEXES
        elif op_parser.op_equals(StoreOperation.LIST_INDEXES):
            result = []
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            key = op_parser.get_key()
            db_key = get_db_key_from_key(key)
            if db_key in data:
                value = data[db_key]
                result = build_item_from_value(
                    processor=processor, value=value, include_value=True
                )
            else:
                raise NotFoundError
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            document = op_parser.get_value()
            key = op_parser.get_key()
            where = op_parser.get_where()
            document = processor.add_embed_fields(document, key)
            exists = op_parser.get_where_exists()
            db_key = get_db_key_from_value(document)
            with self._lock:
                if where is None:
                    data[db_key] = document
                elif exists is False:
                    if db_key in data:
                        raise PreconditionFailedError
                    data[db_key] = document
                elif exists is True:
                    if db_key not in data:
                        raise PreconditionFailedError
                    data[db_key] = document
                elif where is not None:
                    current_document = None
                    if db_key in data:
                        current_document = data[db_key]
                    if not QueryProcessor.eval_expr(
                        current_document, where, processor.resolve_field
                    ):
                        raise PreconditionFailedError
                    data[db_key] = document
            result = build_item_from_value(processor=processor, value=document)
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            key = op_parser.get_key()
            where = op_parser.get_where()
            returning = op_parser.get_returning()
            set = op_parser.get_set()
            db_key = get_db_key_from_key(key)
            if processor.needs_local_etag():
                etag = processor.generate_etag()
                set = processor.add_etag_update(set, etag)
            current_document = None
            updated_document = None
            with self._lock:
                if db_key in data:
                    current_document = data[db_key]
                if where is None:
                    if current_document is None:
                        raise NotFoundError
                else:
                    if not QueryProcessor.eval_expr(
                        current_document, where, processor.resolve_field
                    ):
                        raise PreconditionFailedError
                updated_document = QueryProcessor.update_item(
                    current_document, set, processor.resolve_field
                )
                data[db_key] = updated_document
            if returning is True:
                result = build_item_from_value(
                    processor=processor,
                    value=updated_document,
                    include_value=True,
                )
            elif returning is False:
                result = build_item_from_value(
                    processor=processor,
                    value=current_document,
                    include_value=True,
                )
            else:
                result = build_item_from_value(
                    processor=processor,
                    value=updated_document,
                    include_value=False,
                )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            key = op_parser.get_key()
            where = op_parser.get_where()
            db_key = get_db_key_from_key(key)
            current_document = None
            with self._lock:
                if db_key in data:
                    current_document = data[db_key]
                if where is None:
                    if current_document is None:
                        raise NotFoundError
                else:
                    if not QueryProcessor.eval_expr(
                        current_document, where, processor.resolve_field
                    ):
                        raise PreconditionFailedError
                data.pop(db_key)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = list(data.values())
            nresult = QueryProcessor.query_items(
                items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_field,
            )
            items = [
                build_item_from_value(
                    processor=processor, value=item, include_value=True
                )
                for item in nresult
            ]
            result = build_query_result(items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = list(data.values())
            result = QueryProcessor.count_items(
                items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_field,
            )
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for op_parser in op_parsers:
                if op_parser.op_equals(StoreOperation.PUT):
                    document = op_parser.get_value()
                    key = op_parser.get_key()
                    document = processor.add_embed_fields(document, key)
                    db_key = get_db_key_from_value(document)
                    with self._lock:
                        data[db_key] = document
                    result.append(
                        build_item_from_value(
                            processor=processor, value=document
                        )
                    )
                elif op_parser.op_equals(StoreOperation.DELETE):
                    key = op_parser.get_key()
                    db_key = get_db_key_from_key(key)
                    with self._lock:
                        if db_key in data:
                            data.pop(db_key)
                    result.append(None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            fail = False
            with self._lock:
                for i in range(0, len(op_parsers)):
                    op_parser = op_parsers[i]
                    processor = collections[i].processor
                    data = collections[i].data
                    if op_parser.op_equals(StoreOperation.PUT):
                        key = op_parser.get_key()
                        where = op_parser.get_where()
                        exists = op_parser.get_where_exists()
                        document = op_parser.get_value()
                        db_key = get_db_key_from_value(document)
                        if where is not None:
                            if exists is False and db_key in data:
                                fail = True
                                break
                            elif exists is True and db_key not in data:
                                fail = True
                                break
                            else:
                                current_document = None
                                if db_key in data:
                                    current_document = data[db_key]
                                if not QueryProcessor.eval_expr(
                                    current_document,
                                    where,
                                    processor.resolve_field,
                                ):
                                    fail = True
                                    break
                    elif op_parser.op_equals(
                        StoreOperation.UPDATE
                    ) or op_parser.op_equals(StoreOperation.DELETE):
                        key = op_parser.get_key()
                        where = op_parser.get_where()
                        db_key = get_db_key_from_key(key)
                        if db_key not in data:
                            fail = True
                            break
                        current_document = data[db_key]
                        if where is not None:
                            if not QueryProcessor.eval_expr(
                                current_document,
                                where,
                                processor.resolve_field,
                            ):
                                fail = True
                                break
                if fail:
                    raise ConflictError
                for i in range(0, len(op_parsers)):
                    op_parser = op_parsers[i]
                    processor = collections[i].processor
                    data = collections[i].data
                    if op_parser.op_equals(StoreOperation.PUT):
                        key = op_parser.get_key()
                        document = op_parser.get_value()
                        document = processor.add_embed_fields(document, key)
                        db_key = get_db_key_from_value(document)
                        data[db_key] = document
                        result.append(
                            build_item_from_value(
                                processor=processor, value=document
                            )
                        )
                    elif op_parser.op_equals(StoreOperation.UPDATE):
                        key = op_parser.get_key()
                        returning = op_parser.get_returning()
                        set = op_parser.get_set()
                        db_key = get_db_key_from_key(key)
                        if processor.needs_local_etag():
                            etag = processor.generate_etag()
                            set = processor.add_etag_update(set, etag)
                        current_document = data[db_key]
                        updated_document = QueryProcessor.update_item(
                            current_document, set, processor.resolve_field
                        )
                        data[db_key] = updated_document
                        if returning is True:
                            result.append(
                                build_item_from_value(
                                    processor=processor,
                                    value=updated_document,
                                    include_value=True,
                                )
                            )
                        elif returning is False:
                            result.append(
                                build_item_from_value(
                                    processor=processor,
                                    value=current_document,
                                    include_value=True,
                                )
                            )
                        else:
                            result.append(
                                build_item_from_value(
                                    processor=processor,
                                    value=updated_document,
                                    include_value=False,
                                )
                            )
                    elif op_parser.op_equals(StoreOperation.DELETE):
                        key = op_parser.get_key()
                        db_key = get_db_key_from_key(key)
                        data.pop(db_key)
                        result.append(None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        else:
            return super().run(operation)
        return Response(result=result)


class MemoryCollection:
    data: dict[str, Any]
    processor: ItemProcessor

    def __init__(
        self,
        data: dict[str, Any],
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ):
        self.data = data
        self.processor = ItemProcessor(
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )



"""
Secret Store on Google Secret Manager.
"""

__all__ = ["GoogleSecretManager"]

from datetime import datetime
from typing import Any

from google.api_core.exceptions import AlreadyExists, NotFound

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import NotFoundError, PreconditionFailedError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)


class GoogleSecretManager(StoreProvider):
    project_id: str
    service_account_info: str | None
    service_account_file: str | None
    access_token: str | None
    nparams: dict[str, Any]

    _credentials: Any
    _client: Any
    _aclient: Any
    _processor: ItemProcessor

    def __init__(
        self,
        project_id: str,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            project_id:
                Google project id.
            service_account_info:
                Google service account info with serialized credentials.
            service_account_file:
                Google service account file with credentials.
            access_token:
                Google access token.
            nparams:
                Native params to Google Secret Manager client.
        """
        self.project_id = project_id
        self.service_account_info = service_account_info
        self.service_account_file = service_account_file
        self.access_token = access_token
        self.nparams = nparams

        self._credentials = None
        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        if self._credentials is not None:
            from verse.internal.storage_common._google import get_credentials

            self._credentials = get_credentials(
                self.service_account_info,
                self.service_account_file,
                self.access_token,
            )

        from google.cloud.secretmanager import SecretManagerServiceClient

        if self._credentials is not None:
            self._client = SecretManagerServiceClient(
                credentials=self._credentials
            )
        else:
            self._client = SecretManagerServiceClient()

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        if self._credentials is not None:
            from verse.internal.storage_common._google import get_credentials

            self._credentials = get_credentials(
                self.service_account_info,
                self.service_account_file,
                self.access_token,
            )

        from google.cloud.secretmanager import SecretManagerServiceAsyncClient

        if self._credentials is not None:
            self._aclient = SecretManagerServiceAsyncClient(
                credentials=self._credentials
            )
        else:
            self._aclient = SecretManagerServiceAsyncClient()

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client, self.project_id),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._aclient,
            AsyncClientHelper(self._aclient, self.project_id),
        )
        if ncall is None:
            return await super().arun(operation)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        project_id = self.project_id
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            ver = "latest" if version is None else version
            path = f"projects/{project_id}/secrets/{id}/versions/{ver}"
            args: dict[str, Any] = {"name": path}
            call = NCall(
                client.access_secret_version,
                args,
                nargs,
                {NotFound: NotFoundError},
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            path = f"projects/{project_id}/secrets/{id}"
            args = {"name": path}
            call = NCall(
                client.get_secret,
                args,
                nargs,
                {NotFound: NotFoundError},
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"id": id, "nargs": nargs}
            call = NCall(helper.get_versions, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata = op_parser.get_metadata()
            exists = op_parser.get_where_exists()
            args = {
                "id": id,
                "value": value,
                "metadata": metadata if metadata is not None else {},
                "where_exists": exists,
                "nargs": nargs,
            }
            call = NCall(helper.put, args)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            path = f"projects/{project_id}/secrets/{id}"
            args = {
                "request": {
                    "parent": path,
                    "payload": {"data": str(value).encode("utf_8")},
                }
            }
            call = NCall(
                client.add_secret_version,
                args,
                nargs,
                {NotFound: NotFoundError},
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            path = f"projects/{project_id}/secrets/{id}"
            args = {
                "request": {
                    "secret": {"name": path, "labels": metadata},
                    "update_mask": {"paths": ["labels"]},
                }
            }
            call = NCall(
                client.update_secret,
                args,
                nargs,
                {NotFound: NotFoundError},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            path = f"projects/{project_id}/secrets/{id}"
            args = {"name": path}
            call = NCall(
                client.delete_secret, args, nargs, {NotFound: NotFoundError}
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            path = f"projects/{project_id}"
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            path = f"projects/{project_id}"
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id, version=nresult.name.split("/")[-1]),
                value=nresult.payload.data.decode("utf-8"),
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id),
                metadata=dict(nresult.labels),
                properties=SecretProperties(
                    created_time=Helper.convert_to_datetime(
                        nresult.create_time.timestamp()
                    )
                ),
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            versions: list = []
            for item in nresult:
                versions.append(
                    SecretVersion(
                        version=item.name.split("/")[-1],
                        created_time=Helper.convert_to_datetime(
                            item.create_time.timestamp()
                        ),
                    )
                )
            return SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id, version=nresult.name.split("/")[-1])
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id, version=nresult.name.split("/")[-1])
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            result = SecretItem(
                key=SecretKey(id=id),
                metadata=metadata if metadata is not None else {},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.name.split("/")[-1]),
                        metadata=dict(item.labels),
                        properties=SecretProperties(
                            created_time=Helper.convert_to_datetime(
                                item.create_time.timestamp()
                            )
                        ),
                    )
                )
            items = QueryProcessor.query_items(
                items=items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_root_field,
            )
            result = SecretList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = []
            for item in nresult:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.name.split("/")[-1]),
                        metadata=dict(item.labels),
                        properties=SecretProperties(
                            created_time=Helper.convert_to_datetime(
                                item.create_time.timestamp()
                            )
                        ),
                    )
                )
            result = QueryProcessor.count_items(
                items=items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_root_field,
            )
        return result


class Helper:
    @staticmethod
    def convert_to_datetime(timestamp):
        return datetime.fromtimestamp(timestamp).astimezone()


class ClientHelper:
    client: Any
    project_id: str

    def __init__(self, client: Any, project_id: str):
        self.client = client
        self.project_id = project_id

    def get_versions(self, id: str, nargs: Any):
        path = f"projects/{self.project_id}/secrets/{id}"
        return NCall(
            self.client.list_secret_versions,
            {"parent": path},
            nargs,
            {NotFound: NotFoundError},
        ).invoke()

    def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool | None,
        nargs: Any,
    ) -> Any:
        error = None
        if where_exists is None or where_exists is False:
            (response, error) = NCall(
                self.client.create_secret,
                {
                    "request": {
                        "parent": f"projects/{self.project_id}",
                        "secret_id": id,
                        "secret": {
                            "replication": {"automatic": {}},
                            "labels": metadata,
                        },
                    }
                },
                nargs,
                {
                    AlreadyExists: (
                        PreconditionFailedError
                        if where_exists is False
                        else None
                    )
                },
            ).invoke(return_error=True)
        path = f"projects/{self.project_id}/secrets/{id}"
        response = NCall(
            self.client.add_secret_version,
            {
                "request": {
                    "parent": path,
                    "payload": {"data": value.encode("utf_8")},
                }
            },
            nargs,
            {NotFound: PreconditionFailedError},
        ).invoke()
        if error is not None or where_exists is True:
            NCall(
                self.client.update_secret,
                {
                    "request": {
                        "secret": {"name": path, "labels": metadata},
                        "update_mask": {"paths": ["labels"]},
                    }
                },
                nargs,
            ).invoke()
        return response

    def query(self, nargs: Any):
        path = f"projects/{self.project_id}"
        response = NCall(
            self.client.list_secrets,
            {"request": {"parent": path}},
            nargs,
        ).invoke()
        return response

    def close(self, nargs: Any) -> Any:
        pass


class AsyncClientHelper:
    client: Any
    project_id: str

    def __init__(self, client: Any, project_id: str):
        self.client = client
        self.project_id = project_id

    async def get_versions(self, id: str, nargs: Any):
        path = f"projects/{self.project_id}/secrets/{id}"
        response = await NCall(
            self.client.list_secret_versions,
            {"parent": path},
            nargs,
            {NotFound: NotFoundError},
        ).ainvoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool,
        nargs: Any,
    ) -> Any:
        error = None
        if where_exists is None or where_exists is False:
            (response, error) = await NCall(
                self.client.create_secret,
                {
                    "request": {
                        "parent": f"projects/{self.project_id}",
                        "secret_id": id,
                        "secret": {
                            "replication": {"automatic": {}},
                            "labels": metadata,
                        },
                    }
                },
                nargs,
                {
                    AlreadyExists: (
                        PreconditionFailedError
                        if where_exists is False
                        else None
                    )
                },
            ).ainvoke(return_error=True)
        path = f"projects/{self.project_id}/secrets/{id}"
        response = await NCall(
            self.client.add_secret_version,
            {
                "request": {
                    "parent": path,
                    "payload": {"data": value.encode("utf_8")},
                }
            },
            nargs,
            {NotFound: PreconditionFailedError},
        ).ainvoke()
        if error is not None or where_exists is True:
            await NCall(
                self.client.update_secret,
                {
                    "request": {
                        "secret": {"name": path, "labels": metadata},
                        "update_mask": {"paths": ["labels"]},
                    }
                },
                nargs,
            ).ainvoke()
        return response

    async def query(self, nargs: Any):
        path = f"projects/{self.project_id}"
        response = await NCall(
            self.client.list_secrets,
            {"request": {"parent": path}},
            nargs,
        ).ainvoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def close(self, nargs: Any):
        pass



"""
Config Store on AWS Parameter Store.
"""

__all__ = ["AWSParameterStore"]

from typing import Any

import boto3

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import QueryArgs, get_query_args, normalize_label
from .._models import ConfigItem, ConfigKey, ConfigList, ConfigProperties


class AWSParameterStore(StoreProvider):
    region_name: str | None
    profile_name: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    tier: str
    nparams: dict[str, Any]

    _client: Any
    _processor: ItemProcessor

    def __init__(
        self,
        region_name: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        tier: str = "Standard",
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            region_name:
                AWS region name.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key id.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            tier:
                AWS Parameter Store storage tier, defaults to "Standard".
            nparams:
                Native parameters to boto3 client.
        """
        self.region_name = region_name
        self.profile_name = profile_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.tier = tier
        self.nparams = nparams

        self._client = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        session = None
        if self.profile_name is not None:
            session = boto3.session.Session(
                profile_name=self.profile_name,
                **self.nparams,
            )
        elif (
            self.aws_access_key_id is not None
            and self.aws_secret_access_key is not None
        ):
            session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                **self.nparams,
            )
        else:
            session = boto3.session.Session(
                **self.nparams,
            )

        self._client = session.client(
            service_name="ssm",
            region_name=self.region_name,
            **self.nparams,
        )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client, self.tier),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            args = {
                "id": id,
                "label": label,
                "nargs": nargs,
            }
            call = NCall(helper.get, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            label = normalize_label(op_parser.get_label())
            args = {
                "id": id,
                "label": label,
                "value": value,
                "nargs": nargs,
            }
            call = NCall(helper.put, args)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            args = {"id": id, "label": label, "nargs": nargs}
            call = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            result = ConfigItem(
                key=ConfigKey(
                    id=nresult["Parameter"]["Name"],
                    label=nresult["Parameter"]["Selector"][1:],
                ),
                value=nresult["Parameter"]["Value"],
                properties=ConfigProperties(
                    updated_time=nresult["Parameter"]["LastModifiedDate"]
                ),
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            label = normalize_label(op_parser.get_label())
            result = ConfigItem(
                key=ConfigKey(id=id, label=label),
                value=value,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult:
                if len(item["Labels"]) > 0:
                    items.append(
                        ConfigItem(
                            key=ConfigKey(
                                id=item["Name"], label=item["Labels"][0]
                            ),
                            value=item["Value"],
                            properties=ConfigProperties(
                                updated_time=item["LastModifiedDate"]
                            ),
                        )
                    )
            items = sorted(items, key=lambda x: (x.key.label, x.key.id))
            result = ConfigList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            count = 0
            for item in nresult:
                if len(item["Labels"]) > 0:
                    count = count + 1
            result = count
        return result


class OperationConverter:
    pass


class ClientHelper:
    client: Any
    tier: str

    def __init__(self, client: Any, tier: str):
        self.client = client
        self.tier = tier

    def get(self, id: str, label: str, nargs: Any) -> Any:
        ex = self.client.exceptions
        args: dict[str, Any] = {"Name": f"{id}:{label}"}
        return NCall(
            self.client.get_parameter,
            args,
            nargs,
            {
                ex.ParameterNotFound: NotFoundError,
                ex.ParameterVersionNotFound: NotFoundError,
            },
        ).invoke()

    def put(
        self,
        id: str,
        label: str,
        value: str,
        nargs: Any,
    ) -> Any:
        args = {
            "Name": id,
            "Value": value,
            "Type": "String",
            "Overwrite": True,
            "Tier": self.tier,
        }
        res = NCall(self.client.put_parameter, args, nargs).invoke()
        version = res["Version"]
        args = {"Name": id, "ParameterVersion": version, "Labels": [label]}
        NCall(self.client.label_parameter_version, args, nargs).invoke()
        return res

    def delete(self, id: str, label: str, nargs: Any) -> Any:
        res = self.get(id, label, nargs)
        version = res["Parameter"]["Version"]
        res = NCall(
            self.client.unlabel_parameter_version,
            {
                "Name": id,
                "ParameterVersion": version,
                "Labels": [label],
            },
            nargs,
        ).invoke()
        if len(res["InvalidLabels"]) > 0:
            raise BadRequestError("Could not delete. Try again!")

        # Removing the label is async.
        # Check if no versions have labels other than
        # the current label and delete parameter.
        no_labels = True
        versions = self._list_versions(id, nargs)
        for item in versions:
            if len(item["Labels"]) == 1:
                if item["Labels"][0] == label and item["Version"] == version:
                    pass
                else:
                    no_labels = False
            elif len(item["Labels"]) > 1:
                no_labels = False
        if no_labels:
            res = NCall(
                self.client.delete_parameter,
                {"Name": id},
                nargs,
            ).invoke()
        return None

    def query(self, query_args: QueryArgs, nargs: Any) -> Any:
        return self._query_parameters(
            query_args.id_filter, query_args.label_filter, nargs
        )

    def close(self, nargs: Any) -> Any:
        pass

    def _query_parameters(
        self, id_filter: str | None, label_filter: str | None, nargs: Any
    ) -> Any:
        parameters = []
        next_token = None
        while True:
            args = None
            if id_filter:
                args = {
                    "ParameterFilters": [
                        {
                            "Key": "Name",
                            "Option": "BeginsWith",
                            "Values": [id_filter],
                        }
                    ]
                }
            if next_token is not None:
                if args is None:
                    args = {"NextToken": next_token}
                else:
                    args = args | {"NextToken": next_token}
            res = NCall(
                self.client.describe_parameters,
                args,
                nargs,
            ).invoke()
            parameters.extend(res["Parameters"])
            if "NextToken" in res and res["NextToken"] is not None:
                next_token = res["NextToken"]
            else:
                next_token = None
            if next_token is None:
                break
        result = []
        for parameter in parameters:
            versions = self._list_versions(parameter["Name"], nargs)
            if label_filter is not None:
                for item in versions:
                    if label_filter in item["Labels"]:
                        result.append(item)
            else:
                result.extend(versions)
        return result

    def _list_versions(self, id: str, nargs: Any) -> Any:
        versions = []
        next_token = None
        while True:
            args = {"Name": id}
            if next_token is not None:
                args["NextToken"] = next_token
            res = NCall(
                self.client.get_parameter_history,
                args,
                nargs,
            ).invoke()
            versions.extend(res["Parameters"])
            if "NextToken" in res and res["NextToken"] is not None:
                next_token = res["NextToken"]
            else:
                next_token = None
            if next_token is None:
                break
        return versions

    # get_parameters_by_path has strict begins with semantics.
    # not using for now.
    def _list(self, query_args: QueryArgs, nargs: Any):
        parameters = []
        id_filter = query_args.id_filter
        label_filter = query_args.label_filter
        next_token = None
        while True:
            args: dict = {"Path": id_filter}
            if label_filter is not None:
                args = args | {
                    "ParameterFilters": {
                        "Key": "Label",
                        "Values": [label_filter],
                    }
                }
            if next_token is not None:
                args["NextToken"] = next_token
            res = NCall(
                self.client.get_parameters_by_path,
                args,
                nargs,
            ).invoke()
            parameters.extend(res["Parameters"])
            if "NextToken" in res and res["NextToken"] is not None:
                next_token = res["NextToken"]
            else:
                next_token = None
            if next_token is None:
                break
        return parameters
